// RGB方式で色を管理するクラス
class RGBColor {
 public:
  uint8_t red;
  uint8_t green;
  uint8_t blue;

  // コンストラクタ
  //   指定が無ければ0で初期化
  RGBColor() : red(0), green(0), blue(0) {}

  // コンストラクタ
  //   0 - 255の整数で色を指定
  RGBColor(uint8_t red, uint8_t green, uint8_t blue) : red(red), green(green), blue(blue) {}

  // 0.0 - 100.0の小数で色を指定
  void setFloat(float r, float g, float bb) {
    red = (int)(r * 2.55);
    green = (int)(g * 2.55);
    blue = (int)(bb * 2.55);
  }
};

// HSV方式で色を管理するクラス
class HSVColor {
 public:
  float hue = 0.0f;  // 色相     0.0 - 360.0
  float sat = 0.0f;  // 色の濃さ 0.0 - 100.0
  float val = 0.0f;  // 明るさ   0.0 - 100.0

  // コンストラクタ
  //   指定が無ければ0で初期化
  HSVColor() : hue(0.0f), sat(0.0f), val(0.0f) {}

  // コンストラクタ
  //   初期値の色を指定
  HSVColor(float hue, float sat, float val) : hue(hue), sat(sat), val(val) {}

  // RGB形式に変換
  RGBColor toRGB() {
    float r = val;
    float g = val;
    float b = val;

    while (hue >= 360.0f) hue -= 360.0f;

    if (sat > 0.0f) {
      float hd = hue / 60.0f;
      int i = (int)hd;
      float f = hd - (float)i;
      switch (i) {
        case 0:
          g *= 1.0f - sat * (1.0f - f) / 100.0f;
          b *= 1.0f - sat / 100.0f;
          break;
        case 1:
          r *= 1.0f - sat * f / 100.0f;
          b *= 1.0f - sat / 100.0f;
          break;
        case 2:
          r *= 1.0f - sat / 100.0f;
          b *= 1.0f - sat * (1.0f - f) / 100.0f;
          break;
        case 3:
          r *= 1.0f - sat / 100.0f;
          g *= 1.0f - sat * f / 100.0f;
          break;
        case 4:
          r *= 1.0f - sat * (1.0f - f) / 100.0f;
          g *= 1.0f - sat / 100.0f;
          break;
        case 5:
          g *= 1.0f - sat / 100.0f;
          b *= 1.0f - sat * f / 100.0f;
          break;
      }
    }

    RGBColor rgb;
    rgb.setFloat(r, g, b);
    return rgb;
  }
};

uint8_t sendBuf[150];  // LEDに送信するデータを格納しておくバッファー

// 起動(リセット)後に以下のsetup関数が1回実行される
void setup() {
  // 下準備 使うピンを出力に設定
  pinMode(14, OUTPUT);  // チャネル1 Pin14, PC0
  pinMode(15, OUTPUT);  // チャネル2 Pin15, PC1
  pinMode(16, OUTPUT);  // チャネル3 Pin16, PC2
  pinMode(17, OUTPUT);  // チャネル4 Pin17, PC3
  pinMode(18, OUTPUT);  // チャネル5 Pin18, PC4
  pinMode(8, OUTPUT);   // LED

  HSVColor hsv;  // HSVモデルの色データを作成
  hsv.hue = 0.0f;  // 色相の値 0.0-360.0 単精度浮動小数点数なので末尾にfを付けている
  hsv.sat = 100.0f;  // 彩度の値 0.0-360.0
  hsv.val = 100.0f;  // 明るさの値 0.0-360.0

  // ループ
  // iの値を0から5まで1ずつ増やしながら{}内を実行
  for (uint16_t i = 0; i < 6; i++) {
    RGBColor rgb = hsv.toRGB();     // HSVからRGBに変換
    sendBuf[i * 3] = rgb.green;     // i+1番目のLEDの緑の値
    sendBuf[i * 3 + 1] = rgb.red;   // i+1番目のLEDの赤の値
    sendBuf[i * 3 + 2] = rgb.blue;  // i+1番目のLEDの青の値

    hsv.hue = hsv.hue + 60.0f;  // 次のLEDの色相を60°変化させる
  }

  sendLedData();  // バッファーのデータをアドレサブルRGBデバイスに送信
}

// setup関数の後、loop関数が繰り返し実行される
void loop() {}

// バッファーのデータをアドレサブルRGB LEDに送信する関数
// 中身を変更する必要はない
void sendLedData() {
  uint8_t ch = 0x1F;  // 全チャネルに同じデータを送信
  uint16_t start = 0;
  uint8_t buf;
  uint16_t i = start;
  uint8_t mask = 0x80;
  uint16_t stp = 150;
  noInterrupts();
  while (true) {
    buf = sendBuf[i];
    while (true) {
      if ((buf & mask) == 0) {
        PORTC = ch;
        mask = mask >> 1;
        PORTC = 0;
        if (mask == 0) {
          i++;
          if (i == stp) {
            delayMicroseconds(350);  // 全てのデータ送信後はしばらくLOWにする
            interrupts();
            return;
          }
          mask = 0x80;
          break;
        }
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
      } else {
        PORTC = ch;
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
        mask = mask >> 1;
        if (mask == 0) {
          mask = 0x80;
          i++;
          if (i == stp) {
            PORTC = 0;
            delayMicroseconds(350);  // 全てのデータ送信後はしばらくLOWにする
            interrupts();
            return;
          }
          PORTC = 0;
          break;
        }
        PORTC = 0;
      }
    }
  }
}
